# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.

from pkg_resources import get_distribution

# -- Project information -----------------------------------------------------

project = 'gqlclient'
copyright = '2019, Scott Wiant, Stuart Mumford'
author = 'Scott Wiant, Stuart Mumford'

release = get_distribution('gqlclient').version
version = '.'.join(release.split('.')[:2])


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.intersphinx',
    'sphinx_automodapi.automodapi',
    'sphinx_automodapi.smart_resolver',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Extension configuration -------------------------------------------------

automodsumm_inherited_members = True

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/': None,
                       'https://pydantic-docs.helpmanual.io/': None}



"""
Implementation of the Base graphql client to support the asynchronous creation and
execution of graphql queries and mutations
"""
import logging

import aiohttp

from gqlclient.base import GraphQLClientBase


__all__ = ["AsyncGraphQLClient"]


logger = logging.getLogger(__name__)


class AsyncGraphQLClient(GraphQLClientBase):
    """
    Helper class for formatting and executing asynchronous GraphQL queries and mutations

    """

    async def execute_gql_call(self, query: dict) -> dict:
        """
        Executes a GraphQL query or mutation using aiohttp.

        :param query: Dictionary formatted graphql query.

        :return: Dictionary containing the response from the GraphQL endpoint.
        """

        logger.debug(f"Executing graphql call: host={self.gql_uri}")

        async with aiohttp.ClientSession() as session:
            async with session.post(self.gql_uri, data=query) as response:
                result = await response.json()
                if response.status > 299:
                    raise ValueError(
                        f"Server returned invalid response: "
                        f"code=HTTP{response.status}, "
                        f"detail={response.json()} "
                    )
                return result



#! usr/bin/env python3.7
"""
Tests for the graphql_client library
"""

from datetime import datetime
from typing import List

from pydantic import BaseModel

from gqlclient import GraphQLClient


# Graphql Client to test
test_client = GraphQLClient(gql_uri="http://localhost:5000/graphql")


# Test input parameter definition
class ParametersTest(BaseModel):
    str_param: str
    int_param: int
    float_param: float
    str_array_param: List[str]
    str_num_param: List[int]
    date_param: datetime
    optional_param: int = None


# Test response model(s)
class ResponseChildTest(BaseModel):
    child_param_1: str
    child_param_2: str


class ResponseParentTest(BaseModel):
    parent_param_1: str
    parent_param_2: str
    child_object: ResponseChildTest


test_query_base = "query_base"
test_mutation_base = "mutation_base"
test_parameters = ParametersTest(
    str_param="A",
    int_param=1,
    float_param=1.1,
    str_array_param=["A", "B"],
    str_num_param=[1, 2],
    date_param="2010-03-25T14:08:00.000000",
)


def test_query_string_with_parameters():
    """
    Test of query string structure when parameter model is included
    :return: None
    """
    test_query = test_client.get_query(
        query_base=test_query_base,
        query_parameters=test_parameters,
        query_response_cls=ResponseParentTest,
    )
    expected_result = {
        "query": '{query_base(str_param: "A", '
        "int_param: 1, "
        "float_param: 1.1, "
        'str_array_param: ["A", "B"], '
        "str_num_param: [1, 2], "
        'date_param: "2010-03-25T14:08:00")'
        "{parent_param_1, parent_param_2, "
        "child_object { child_param_1 child_param_2 }"
        "} }"
    }
    assert test_query == expected_result


def test_query_string_without_parameters():
    """
    Test of query string structure when parameter model is NOT included
    :return: None
    """
    test_query = test_client.get_query(
        query_base=test_query_base, query_response_cls=ResponseParentTest
    )
    expected_result = {
        "query": "{query_base"
        "{parent_param_1, parent_param_2, "
        "child_object { child_param_1 child_param_2 }"
        "} }"
    }

    assert test_query == expected_result


def test_mutation_string_with_response():
    """
    Test of mutation string structure when response model is included
    :return: None
    """
    test_mutation = test_client.get_mutation(
        mutation_base=test_mutation_base,
        mutation_response_cls=ResponseParentTest,
        mutation_parameters=test_parameters,
    )
    expected_result = {
        "query": "mutation mutation_base "
        "{mutation_base("
        'str_param: "A", '
        "int_param: 1, "
        "float_param: 1.1, "
        'str_array_param: ["A", "B"], '
        "str_num_param: [1, 2], "
        'date_param: "2010-03-25T14:08:00")'
        "{ok, "
        "parent_param_1, parent_param_2, "
        "child_object { child_param_1 child_param_2 }"
        "} }",
        "operationName": "mutation_base",
    }
    assert test_mutation == expected_result


def test_mutation_string_without_response():
    """
    Test of mutation string structure when response model is NOT included
    :return: None
    """
    test_mutation = test_client.get_mutation(
        mutation_base=test_mutation_base, mutation_parameters=test_parameters
    )
    print(test_mutation)
    expected_result = {
        "query": "mutation mutation_base "
        "{mutation_base("
        'str_param: "A", '
        "int_param: 1, "
        "float_param: 1.1, "
        'str_array_param: ["A", "B"], '
        "str_num_param: [1, 2], "
        'date_param: "2010-03-25T14:08:00")'
        "{ok"
        "} }",
        "operationName": "mutation_base",
    }
    print(expected_result)
    assert test_mutation == expected_result



"""
Base class to support the creation of graphql queries and mutations.
"""
from abc import ABC, abstractmethod
import json
import datetime
import logging
from typing import Dict, Optional, Type

from pydantic import BaseModel


__all__ = ["GraphQLClientBase"]


logger = logging.getLogger(__name__)


class DatetimeEncoder(json.JSONEncoder):
    """
    A JSON encoder which encodes datetimes as iso formatted strings.
    """

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat("T")
        return super().default(obj)


class GraphQLClientBase(ABC):
    """
    Abstract class for formatting and executing GraphQL queries and mutations

    :param gql_uri: Fully qualified URI for the graphQL endpoint.
    """

    def __init__(self, gql_uri: str):
        self.gql_host = gql_uri

    @staticmethod
    def _graphql_response_from_model(model_cls: Type[BaseModel]) -> str:
        """
        Generate a GraphQL Response query from the class for the response.

        :param model_cls: Model representing the desired response object from the graphql endpoint.

        :return: Portion of a graphql query which specifies the response object.
        """

        def parse_field(field):
            if hasattr(field.type_, "__fields__"):
                return (
                    field.name,
                    [parse_field(sfield) for sfield in field.type_.__fields__.values()],
                )
            else:
                return field.name

        names = [parse_field(f) for f in model_cls.__fields__.values()]
        names = [
            f"{name[0]} {{ {' '.join(name[1])} }}" if isinstance(name, tuple) else name
            for name in names
        ]
        return ", ".join(names)

    @staticmethod
    def _graphql_query_parameters_from_model(model: BaseModel) -> str:
        """
        Generate a GraphQL query parameters from the class for the query.

        :param model: Model representing the query parameters and actual search values (min 1)python

        :return: Portion of a graphql query which specifies the query parameters
        """

        # Only create query parameters for those with values using graphql syntax mask for the query parameters
        parameters = ", ".join(
            [f"{k}: {json.dumps(v, cls=DatetimeEncoder)}" for k, v in model if v]
        )

        return parameters

    def get_query(
        self,
        query_base: str,
        query_response_cls: Type[BaseModel],
        query_parameters: Optional[BaseModel] = None,
    ) -> Dict[str, str]:
        """
        Create a GraphQL formatted query string.

        :param query_base: Name of the root type to be queried
        :param query_response_cls: A pydantic model class representing the structure of the response object
        with attributes  corresponding to the Graphql type and attribute names
        :param query_parameters: Optional. Instance of a pydantic model containing attributes corresponding
        to parameter names and values corresponding to the parameter value.

        :return: Dictionary that can be passed as json to the GraphQL API endpoint
        """
        # Construct graphql query
        gql_query = query_base
        if query_parameters:
            gql_query += f"({self._graphql_query_parameters_from_model(query_parameters)})"
        gql_query += f"{{{self._graphql_response_from_model(query_response_cls)}}}"
        return {"query": f"{{{gql_query} }}"}

    def get_mutation(
        self,
        mutation_base: str,
        mutation_parameters: BaseModel,
        mutation_response_cls: Optional[BaseModel] = None,
    ) -> Dict[str, str]:
        """
        Create a GraphQL formatted mutation string.

        :param mutation_base: Name of the root type to be mutated
        :param mutation_parameters: Instance of a pydantic model containing attributes corresponding to
        parameter names and values corresponding to the parameter value.
        :param mutation_response_cls: Optional. A pydantic model class representing the structure of the
        response object with attributes corresponding to the Graphql type and attribute names.

        :return: Dictionary that can be passed as json to the GraphQL API endpoint
        """
        # Construct graphql mutation
        gql_mutation = (
            f"mutation {mutation_base} {{{mutation_base}"
            f"({self._graphql_query_parameters_from_model(mutation_parameters)}){{ok"
        )
        if mutation_response_cls:
            gql_mutation += f", {self._graphql_response_from_model(mutation_response_cls)}"
        gql_mutation += "}"

        return {"query": f"{gql_mutation} }}", "operationName": mutation_base}

    @abstractmethod
    def execute_gql_call(self, query: dict) -> dict:
        """
        Executes a GraphQL query or mutation.

        :param query: Dictionary formatted graphql query

        :return: Dictionary containing the response from the GraphQL endpoint
        """

    def execute_gql_query(
        self,
        query_base: str,
        query_response_cls: Type[BaseModel],
        query_parameters: Optional[BaseModel],
    ) -> dict:
        """
        Executes a graphql query based upon input pydantic models.

        :param query_base: Name of the root type to be queried
        :param query_parameters: Optional. Instance of a pydantic model containing attributes corresponding to
        parameter names and values corresponding to the parameter value.
        :param query_response_cls: A pydantic model class representing the structure of the response
        object with attributes corresponding to the Graphql type and attribute names

        :return: Dictionary containing the response from the GraphQL endpoint
        """
        query = self.get_query(query_base, query_response_cls, query_parameters)
        return self.execute_gql_call(query)

    def execute_gql_mutation(
        self,
        mutation_base: str,
        mutation_parameters: BaseModel,
        mutation_response_cls: Optional[BaseModel] = None,
    ) -> dict:
        """
        Executes a graphql mutation based upon input pydantic models.

        :param mutation_base: Name of the root type to be mutated
        :param mutation_parameters: Instance of a pydantic model containing attributes corresponding to
        parameter names and values corresponding to the parameter value.
        :param mutation_response_cls: Optional. A pydantic model class representing the structure of the
        response object with attributes corresponding to the Graphql type and attribute names.

        :return: Dictionary containing the response from the GraphQL endpoint
        """
        mutation = self.get_mutation(mutation_base, mutation_parameters, mutation_response_cls)
        return self.execute_gql_call(mutation)



"""
Implementation of the Base graphql client to support the synchronous creation and
execution of graphql queries and mutations
"""
import logging

import requests

from gqlclient.base import GraphQLClientBase


__all__ = ["GraphQLClient"]


logger = logging.getLogger(__name__)


class GraphQLClient(GraphQLClientBase):
    """
    Helper class for formatting and executing synchronous GraphQL queries and mutations
    """

    def execute_gql_call(self, query: dict) -> dict:
        """
        Executes a GraphQL query or mutation using requests.

        :param query: Dictionary formatted graphql query

        :return: Dictionary containing the response from the GraphQL endpoint
        """
        logger.debug(f"Executing graphql call: host={self.gql_uri}")

        response = requests.post(url=self.gql_host, json=query)
        # GraphQL encodes most errors within the body of a 200 response.
        # Non 200 statuses represent something outside of the query parsing and
        # execution having failed.
        if response.status_code > 299:
            raise ValueError(
                f"Server returned invalid response: "
                f"code=HTTP{response.status_code}, "
                f"detail={response.json()} "
            )
        return response.json()
